// zoomasm -- zoom video assembler
// (c) 2019,2020 Claude Heiland-Allen
// SPDX-License-Identifier: AGPL-3.0-or-later

#include "audio.h"

#include "imgui.h"
#include "imfilebrowser.h"

#include <cmath>
#include <cstring>
#include <complex>
#include <filesystem>
#include <thread>

#include <fftw3.h>

#define MA_NO_ENCODING
#define DR_FLAC_IMPLEMENTATION
#include "extras/dr_flac.h"
#define DR_MP3_IMPLEMENTATION
#include "extras/dr_mp3.h"
#define DR_WAV_IMPLEMENTATION
#include "extras/dr_wav.h"
#define STB_VORBIS_HEADER_ONLY
#include "extras/stb_vorbis.c"
#define MINIAUDIO_IMPLEMENTATION
#include "miniaudio.h"
#undef STB_VORBIS_HEADER_ONLY

#include "resource.h"

static inline float dbfstorms(float dbfs)
{
  return std::pow(10.0f, dbfs / 10.0f);
}

static inline float rmstodbfs(float rms)
{
  return 10.0f * std::log10(rms);
}

static inline std::complex<double> normalize(std::complex<double> x)
{
  x /= std::abs(x);
  if (std::isnan(std::real(x)) || std::isnan(std::imag(x)) || std::isinf(std::real(x)) || std::isinf(std::imag(x)))
  {
    return 1;
  }
  return x;
}

static inline std::complex<double> to_std_complex(fftw_complex x)
{
  return std::complex<double>(x[0], x[1]);
}

static inline void to_fftw_complex(std::complex<double> x, fftw_complex y)
{
  y[0] = std::real(x);
  y[1] = std::imag(x);
}

// requests from audio callback to decoder

enum request_type
{
  request_type_time,
  request_type_free,
  request_type_read
};

// player is playing at this timestamp (informational)
struct request_time
{
  double time;
};

// player no longer needs this memory
struct request_free
{
  size_t bytes;
  void *data;
};

// player will need this audio data soon
struct request_read
{
  ma_int64 start; // note: signed
  ma_uint64 frames;
};

struct request
{
  enum request_type type;
  union
  {
    struct request_time time;
    struct request_free free;
    struct request_read read;
  } u;
};

// responses from decoder to audio callback

enum response_type
{
  response_type_volume,
  response_type_speed,
  response_type_time,
  response_type_loop,
  response_type_read,
  response_type_quit
};

// set output volume [0..1]
struct response_volume
{
  double volume;
};

// set playback speed [-10..10] (too fast might glitch)
struct response_speed
{
  double speed;
};

// set playback time (manual seek, might not be instant)
struct response_time
{
  double time;
};

// set playback looping
// FIXME TODO sample-accurate seamless looping
struct response_loop
{
  bool loop;
};

// audio data from decoder (hopefully useful)
struct response_read
{
  ma_int64 start; // note: signed
  ma_uint64 frames;
  size_t bytes;
  float *data;
};

struct response
{
  enum response_type type;
  union
  {
    struct response_volume volume;
    struct response_speed speed;
    struct response_time time;
    struct response_loop loop;
    struct response_read read;
    // quit has no data
  } u;
};

#define AUDIO_RB_BYTES (sizeof(struct request) * sizeof(struct response) * 256)

#define AUDIO_BLOCK_SIZE 4096
#define AUDIO_OVERLAP 4
#define AUDIO_HOP_SIZE 256
#define AUDIO_FFT_SIZE 1024
#define AUDIO_RT_BUFFERS 64

struct audio_rt
{
  bool have_rb_request;
  ma_rb request;
  bool have_rb_response;
  ma_rb response;
  int channels; // 0 for silence (playhead scheduler only)
  std::thread *thread;
  double sample_rate;
  ma_uint64 duration; // frames
  double time; // seconds
  double speed;
  bool loop;
  double volume_start;
  double volume_target;
  double volume_ramp;
  // invariants:
  // - contiguous
  // - sorted
  // - filled with target frame ranges
  // - buffers that are not ready have null data pointer
  // - buffers with null data pointer have been requested
  struct response_read buffer[AUDIO_RT_BUFFERS];
  double *window; // [AUDIO_FFT_SIZE]
  double *fft_in; // [AUDIO_FFT_SIZE]
  fftw_complex *fft_out[2]; // [AUDIO_FFT_SIZE]
  fftw_complex **ifft_in; // [channels][AUDIO_FFT_SIZE]
  double *ifft_out; // [AUDIO_FFT_SIZE]
  float *outbuf; // [channels * AUDIO_HOP_SIZE * AUDIO_OVERLAP]
  fftw_plan fft[2];
  fftw_plan *ifft; // [channels]
  int frame; // [0..AUDIO_HOP_SIZE]
  // error counts
  ma_uint64 errs_request_overrun;
  ma_uint64 errs_response_overrun;
  ma_uint64 errs_read_underrun;
};

static ma_result audio_rt_response_quit(struct audio_rt *art);
static void audio_rt_silence(struct audio_rt *art);

static void audio_rt_uninit(struct audio_rt *art)
{
  if (! art)
  {
    return;
  }
  if (art->thread)
  {
    // signal quit, join thread
    audio_rt_response_quit(art);
    art->thread->join();
    delete art->thread;
    art->thread = nullptr;
  }
  if (art->have_rb_request)
  {
    // free buffers in request ring buffer
    ma_result res = MA_SUCCESS;
    while (res == MA_SUCCESS && ma_rb_pointer_distance(&art->request) >= (ssize_t) sizeof(struct request))
    {
      size_t bytes = sizeof(struct request);
      void *data = nullptr;
      res = ma_rb_acquire_read(&art->request, &bytes, &data);
      if (res == MA_SUCCESS && bytes == sizeof(struct request))
      {
        struct request *req = (struct request *) data;
        switch (req->type)
        {
          case request_type_free:
          {
            if (req->u.free.data)
            {
              free(req->u.free.data);
            }
            break;
          }
          case request_type_read:
          case request_type_time:
            break;
        }
        res = ma_rb_commit_read(&art->request, bytes, data);
        if (res != MA_SUCCESS)
        {
          // FIXME handle error
        }
      }
      else
      {
        res = MA_NO_DATA_AVAILABLE;
        assert(! "split packet?");
      }
    }
    ma_rb_uninit(&art->request);
    art->have_rb_request = false;
  }
  if (art->have_rb_response)
  {
    // free buffers in response ring buffer
    ma_result res = MA_SUCCESS;
    while (res == MA_SUCCESS && ma_rb_pointer_distance(&art->response) >= (ssize_t) sizeof(struct response))
    {
      size_t bytes = sizeof(struct response);
      void *data = nullptr;
      res = ma_rb_acquire_read(&art->response, &bytes, &data);
      if (res == MA_SUCCESS && bytes == sizeof(struct response))
      {
        struct response *resp = (struct response *) data;
        switch (resp->type)
        {
          case response_type_read:
          {
            if (resp->u.read.data)
            {
              free(resp->u.read.data);
            }
            break;
          }
          case response_type_speed:
          case response_type_time:
          case response_type_volume:
          case response_type_loop:
          case response_type_quit:
            break;
        }
        res = ma_rb_commit_read(&art->response, bytes, data);
        if (res != MA_SUCCESS)
        {
          // FIXME handle error
        }
      }
      else
      {
        res = MA_NO_DATA_AVAILABLE;
        assert(! "split packet?");
      }
    }
    ma_rb_uninit(&art->response);
    art->have_rb_response = false;
  }
  for (int b = 0; b < AUDIO_RT_BUFFERS; ++b)
  {
    if (art->buffer[b].data)
    {
      free(art->buffer[b].data);
      art->buffer[b].data = nullptr;
    }
  }
#define F(P) \
  if (art->P) \
  { \
    fftw_free(art->P); \
    art->P = nullptr; \
  }
  F(window)
  F(fft_in)
  F(fft_out[0])
  F(fft_out[1])
  if (art->ifft_in)
  {
    for (int c = 0; c < art->channels; ++c)
    {
      F(ifft_in[c])
    }
    free(art->ifft_in);
    art->ifft_in = nullptr;
  }
  F(ifft_out)
  F(outbuf)
#undef F
#define F(P) \
  if (art->P) \
  { \
    fftw_destroy_plan(art->P); \
    art->P = nullptr; \
  }
  F(fft[0])
  F(fft[1])
  if (art->ifft)
  {
    for (int c = 0; c < art->channels; ++c)
    {
      F(ifft[c])
    }
    free(art->ifft);
    art->ifft = nullptr;
  }
#undef F
}

static ma_result audio_rt_init(struct audio_rt *art, int channels, double sample_rate, ma_uint64 duration)
{
  std::memset(art, 0, sizeof(*art));
  ma_result res = MA_SUCCESS;
  res = ma_rb_init(AUDIO_RB_BYTES, nullptr, nullptr, &art->request);
  if (res != MA_SUCCESS)
  {
    return res;
  }
  art->have_rb_request = true;
  res = ma_rb_init(AUDIO_RB_BYTES, nullptr, nullptr, &art->response);
  if (res != MA_SUCCESS)
  {
    ma_rb_uninit(&art->request);
    art->have_rb_request = false;
    return res;
  }
  art->have_rb_response = true;
  art->channels = channels;
  art->sample_rate = sample_rate;
  art->duration = duration;
  art->frame = 0;
  art->time = 0;
  art->speed = 0;
  art->volume_start = 0;
  art->volume_target = 1;
  art->volume_ramp = 1;
  if (channels == 0)
  {
    art->thread = new std::thread(audio_rt_silence, art);
    if (! art->thread)
    {
      res = MA_OUT_OF_MEMORY;
    }
  }
  else
  {
    art->window = fftw_alloc_real(AUDIO_FFT_SIZE);
    if (! art->window)
    {
      res = MA_OUT_OF_MEMORY;
    }
    double gain = 0.5 / std::sqrt(AUDIO_FFT_SIZE);
    for (int i = 0; i < AUDIO_FFT_SIZE; ++i)
    {
      art->window[i] = gain * (1 - std::cos(i * 6.283185307179586 / AUDIO_FFT_SIZE));
    }
    art->fft_in = fftw_alloc_real(AUDIO_FFT_SIZE);
    if (! art->fft_in)
    {
      res = MA_OUT_OF_MEMORY;
    }
    art->fft_out[0] = fftw_alloc_complex(AUDIO_FFT_SIZE);
    if (! art->fft_out[0])
    {
      res = MA_OUT_OF_MEMORY;
    }
    art->fft_out[1] = fftw_alloc_complex(AUDIO_FFT_SIZE);
    if (! art->fft_out[1])
    {
      res = MA_OUT_OF_MEMORY;
    }
    art->ifft_in = (fftw_complex **) calloc(1, channels * sizeof(*art->ifft_in));
    if (art->fft_in)
    {
      for (int c = 0; c < channels; ++c)
      {
        art->ifft_in[c] = fftw_alloc_complex(AUDIO_FFT_SIZE);
        if (! art->ifft_in[c])
        {
          res = MA_OUT_OF_MEMORY;
        }
      }
    }
    else
    {
      res = MA_OUT_OF_MEMORY;
    }
    art->ifft_out = fftw_alloc_real(AUDIO_FFT_SIZE);
    if (! art->ifft_out)
    {
      res = MA_OUT_OF_MEMORY;
    }
    art->outbuf = (float *) fftw_malloc(sizeof(*art->outbuf) * channels * AUDIO_FFT_SIZE);
  //std::fprintf(stderr, "%p\n", art->outbuf);
    if (! art->outbuf)
    {
      res = MA_OUT_OF_MEMORY;
    }
    art->ifft = (fftw_plan *) fftw_malloc(sizeof(*art->ifft) * channels);
    if (! art->ifft)
    {
      res = MA_OUT_OF_MEMORY;
    }
    if (res == MA_SUCCESS)
    {
      art->fft[0] = fftw_plan_dft_r2c_1d(AUDIO_FFT_SIZE, art->fft_in, art->fft_out[0], FFTW_PATIENT | FFTW_DESTROY_INPUT);
      if (! art->fft[0])
      {
        res = MA_OUT_OF_MEMORY;
      }
      art->fft[1] = fftw_plan_dft_r2c_1d(AUDIO_FFT_SIZE, art->fft_in, art->fft_out[1], FFTW_PATIENT | FFTW_DESTROY_INPUT);
      if (! art->fft[1])
      {
        res = MA_OUT_OF_MEMORY;
      }
      for (int c = 0; c < channels; ++c)
      {
        art->ifft[c] = fftw_plan_dft_c2r_1d(AUDIO_FFT_SIZE, art->ifft_in[c], art->ifft_out, FFTW_PATIENT | FFTW_PRESERVE_INPUT);
        if (! art->ifft[c])
        {
          res = MA_OUT_OF_MEMORY;
        }
      }
    }
  }
  if (res != MA_SUCCESS)
  {
    audio_rt_uninit(art);
  }
  else
  {
    if (channels > 0)
    {
      std::memset(art->fft_in, 0, AUDIO_FFT_SIZE * sizeof(*art->fft_in));
      std::memset(art->fft_out[0], 0, AUDIO_FFT_SIZE * sizeof(*art->fft_out[0]));
      std::memset(art->fft_out[1], 0, AUDIO_FFT_SIZE * sizeof(*art->fft_out[1]));
      for (int c = 0; c < channels; ++c)
      {
        std::memset(art->ifft_in[c], 0, AUDIO_FFT_SIZE * sizeof(*art->ifft_in[c]));
      }
      std::memset(art->ifft_out, 0, AUDIO_FFT_SIZE * sizeof(*art->ifft_out));
      std::memset(art->outbuf, 0, sizeof(*art->outbuf) * channels * AUDIO_FFT_SIZE);
    }
  }
  return res;
}

static ma_result audio_rt_request_time(struct audio_rt *art, double time)
{
  size_t wbytes = sizeof(struct request);
  void *wbuffer = nullptr;
  ma_result wres = ma_rb_acquire_write(&art->request, &wbytes, &wbuffer);
  if (wres == MA_SUCCESS)
  {
    if (wbytes == sizeof(struct request))
    {
      struct request *req = (struct request *) wbuffer;
      req->type = request_type_time;
      req->u.time.time = time;
      wres = ma_rb_commit_write(&art->request, wbytes, wbuffer);
    }
    else
    {
      wres = MA_NO_DATA_AVAILABLE;
    }
  }
  if (wres != MA_SUCCESS)
  {
    art->errs_request_overrun++;
  }
  return wres;
}

static ma_result audio_rt_request_read(struct audio_rt *art, ma_int64 start, ma_uint64 frames)
{
  size_t wbytes = sizeof(struct request);
  void *wbuffer = nullptr;
  ma_result wres = ma_rb_acquire_write(&art->request, &wbytes, &wbuffer);
  if (wres == MA_SUCCESS)
  {
    if (wbytes == sizeof(struct request))
    {
      struct request *req = (struct request *) wbuffer;
      req->type = request_type_read;
      req->u.read.start = start;
      req->u.read.frames = frames;
      wres = ma_rb_commit_write(&art->request, wbytes, wbuffer);
    }
    else
    {
      wres = MA_NO_DATA_AVAILABLE;
    }
  }
  if (wres != MA_SUCCESS)
  {
    art->errs_request_overrun++;
  }
  return wres;
}

static ma_result audio_rt_request_read_all(struct audio_rt *art)
{
  ma_result res = MA_SUCCESS;
  ma_int64 frame = ((ma_int64) std::floor(art->time * art->sample_rate / AUDIO_BLOCK_SIZE)) * AUDIO_BLOCK_SIZE;
  for (int b = 0; b < AUDIO_RT_BUFFERS; ++b)
  {
    art->buffer[b].start = frame + (b - AUDIO_RT_BUFFERS / 2) * AUDIO_BLOCK_SIZE;
    art->buffer[b].frames = AUDIO_BLOCK_SIZE;
    // FIXME these should already be cleared, assert()?...
    art->buffer[b].bytes = 0;
    art->buffer[b].data = nullptr;
    // request read
    ma_result rres = audio_rt_request_read(art, art->buffer[b].start, art->buffer[b].frames);
    if (rres != MA_SUCCESS)
    {
      res = rres;
    }
  }
  return res;
}

static ma_result audio_rt_request_free(struct audio_rt *art, size_t bytes, void *data)
{
  size_t wbytes = sizeof(struct request);
  void *wbuffer = nullptr;
  ma_result wres = ma_rb_acquire_write(&art->request, &wbytes, &wbuffer);
  if (wres == MA_SUCCESS)
  {
    if (wbytes == sizeof(struct request))
    {
      struct request *req = (struct request *) wbuffer;
      req->type = request_type_free;
      req->u.free.bytes = bytes;
      req->u.free.data = data;
      wres = ma_rb_commit_write(&art->request, wbytes, wbuffer);
    }
    else
    {
      wres = MA_NO_DATA_AVAILABLE;
    }
  }
  if (wres != MA_SUCCESS)
  {
    art->errs_request_overrun++;
  }
  return wres;
}

static ma_result audio_rt_request_free_all(struct audio_rt *art)
{
  ma_result res = MA_SUCCESS;
  for (int b = 0; b < AUDIO_RT_BUFFERS; ++b)
  {
    if (art->buffer[b].data)
    {
      ma_result fres = audio_rt_request_free(art, art->buffer[b].bytes, art->buffer[b].data);
      if (fres != MA_SUCCESS)
      {
        res = fres;
      }
      // FIXME leaks buffer memory if errored above
      art->buffer[b].start = 0;
      art->buffer[b].frames = 0;
      art->buffer[b].bytes = 0;
      art->buffer[b].data = nullptr;
    }
  }
  return res;
}

static ma_result audio_rt_seek_to(struct audio_rt *art, double time)
{
//std::fprintf(stderr, "[ SEEK\n");
  ma_result res = MA_SUCCESS;
  ma_int64 frame = std::floor(time * art->sample_rate);
  double fraction = (time * art->sample_rate) / art->duration;
  if (art->loop)
  {
    frame %= (ma_int64) art->duration;
    frame += (ma_int64) art->duration;
    frame %= (ma_int64) art->duration;
    fraction -= std::floor(fraction);
  }
  else
  {
    frame = std::min(std::max((ma_int64) 0, frame), (ma_int64) art->duration);
    fraction = std::min(std::max(0.0, fraction), 1.0);
  }
  time = fraction * art->duration / art->sample_rate;
//std::fprintf(stderr, "frame %lld\n", frame);
  // check if more data needs to be read
  ma_int64 buffer_start = (frame / AUDIO_BLOCK_SIZE) * AUDIO_BLOCK_SIZE;
  ma_int64 buffer = (buffer_start - art->buffer[0].start) / AUDIO_BLOCK_SIZE;
  ma_int64 offset = buffer - AUDIO_RT_BUFFERS / 2;
//std::fprintf(stderr, "offset %lld\n", offset);
  if (std::abs(offset) >= AUDIO_RT_BUFFERS / 2)
  {
//std::fprintf(stderr, "free all / read all\n");
    // free buffers
    audio_rt_request_free_all(art);
    // update time
    art->time = time;
    // update buffer targets
    audio_rt_request_read_all(art);
  }
  else if (offset > 0)
  {
    // seeking forwards
    // update time
    art->time = time;
    // free old buffers
    for (ma_int64 b = 0; b < offset; ++b)
    {
      ma_result fres = audio_rt_request_free(art, art->buffer[b].bytes, art->buffer[b].data);
      art->buffer[b].start = 0;
      art->buffer[b].frames = 0;
      art->buffer[b].bytes = 0;
      art->buffer[b].data = nullptr;
      if (fres != MA_SUCCESS)
      {
        res = fres;
      }
    }
    // move existing buffers down
//std::fprintf(stderr, "memmove %p %p %lld\n",&art->buffer[0], &art->buffer[offset], sizeof(art->buffer[0]) * (AUDIO_RT_BUFFERS - offset));
    std::memmove(&art->buffer[0], &art->buffer[offset], sizeof(art->buffer[0]) * (AUDIO_RT_BUFFERS - offset));
    // request remaining buffers
    for (ma_int64 b = AUDIO_RT_BUFFERS - offset; b < AUDIO_RT_BUFFERS; ++b)
    {
      ma_int64 start = art->buffer[b - 1].start + AUDIO_BLOCK_SIZE;
      ma_uint64 frames = AUDIO_BLOCK_SIZE;
      art->buffer[b].start = start;
      art->buffer[b].frames = frames;
      art->buffer[b].bytes = 0;
      art->buffer[b].data = nullptr;
      ma_result rres = audio_rt_request_read(art, start, frames);
      if (rres != MA_SUCCESS)
      {
        res = rres;
      }
    }
  }
  else if (offset < 0)
  {
    // seeking backwards
    // update time
    art->time = time;
    // free old buffers
    for (ma_int64 b = AUDIO_RT_BUFFERS + offset; b < AUDIO_RT_BUFFERS; ++b)
    {
      ma_result fres = audio_rt_request_free(art, art->buffer[b].bytes, art->buffer[b].data);
      art->buffer[b].start = 0;
      art->buffer[b].frames = 0;
      art->buffer[b].bytes = 0;
      art->buffer[b].data = nullptr;
      if (fres != MA_SUCCESS)
      {
        res = fres;
      }
    }
    // move existing buffers up
//std::fprintf(stderr, "memmove %p %p %lld\n", &art->buffer[-offset], &art->buffer[0], sizeof(art->buffer[0]) * (AUDIO_RT_BUFFERS - offset));
    std::memmove(&art->buffer[-offset], &art->buffer[0], sizeof(art->buffer[0]) * (AUDIO_RT_BUFFERS + offset));
    // request remaining buffers
    for (ma_int64 b = -offset - 1; b >= 0; --b)
    {
      ma_int64 start = art->buffer[b + 1].start - AUDIO_BLOCK_SIZE;
      ma_uint64 frames = AUDIO_BLOCK_SIZE;
      art->buffer[b].start = start;
      art->buffer[b].frames = frames;
      art->buffer[b].bytes = 0;
      art->buffer[b].data = nullptr;
      ma_result rres = audio_rt_request_read(art, start, frames);
      if (rres != MA_SUCCESS)
      {
        res = rres;
      }
    }
  }
  else // offset == 0
  {
//std::fprintf(stderr, "nop\n");
    // nothing to do
    art->time = time;
  }
//std::fprintf(stderr, "] SEEK\n");
  return res;
}

static ma_result audio_rt_response_read(struct audio_rt *art, ma_int64 start, ma_uint64 frames, size_t bytes, float *data)
{
  size_t wbytes = sizeof(struct response);
  void *wbuffer = nullptr;
  ma_result wres = ma_rb_acquire_write(&art->response, &wbytes, &wbuffer);
  if (wres == MA_SUCCESS)
  {
    if (wbytes == sizeof(struct response))
    {
      struct response *resp = (struct response *) wbuffer;
      resp->type = response_type_read;
      resp->u.read.start = start;
      resp->u.read.frames = frames;
      resp->u.read.bytes = bytes;
      resp->u.read.data = data;
      wres = ma_rb_commit_write(&art->response, wbytes, wbuffer);
    }
    else
    {
      wres = MA_NO_DATA_AVAILABLE;
    }
  }
  if (wres != MA_SUCCESS)
  {
    art->errs_response_overrun++;
  }
  return wres;
}

static ma_result audio_rt_response_quit(struct audio_rt *art)
{
  size_t wbytes = sizeof(struct response);
  void *wbuffer = nullptr;
  ma_result wres = ma_rb_acquire_write(&art->response, &wbytes, &wbuffer);
  if (wres == MA_SUCCESS)
  {
    if (wbytes == sizeof(struct response))
    {
      struct response *resp = (struct response *) wbuffer;
      resp->type = response_type_quit;
      wres = ma_rb_commit_write(&art->response, wbytes, wbuffer);
    }
    else
    {
      wres = MA_NO_DATA_AVAILABLE;
    }
  }
  if (wres != MA_SUCCESS)
  {
    art->errs_response_overrun++;
  }
  return wres;
}


static void audio_rt_silence(struct audio_rt *art)
{
  if (! art)
  {
    return;
  }
  bool running = true;
  while (running)
  {

    // handle messages from decoder
    ma_result rres = MA_SUCCESS;
    while (running && rres == MA_SUCCESS && ma_rb_pointer_distance(&art->response) >= (ssize_t) sizeof(struct response))
    {
      size_t rbytes = sizeof(struct response);
      void *rbuffer = nullptr;
      rres = ma_rb_acquire_read(&art->response, &rbytes, &rbuffer);
      if (rres == MA_SUCCESS)
      {
        if (rbytes == sizeof(struct response))
        {
          struct response *resp = (struct response *) rbuffer;
          switch (resp->type)
          {
            case response_type_volume:
            {
              double volume = art->volume_target + art->volume_ramp * (art->volume_start - art->volume_target);
              art->volume_start = volume;
              art->volume_target = resp->u.volume.volume;
              art->volume_ramp = 1;
              break;
            }
            case response_type_speed:
            {
              art->speed = resp->u.speed.speed; // FIXME TODO fade
              break;
            }
            case response_type_loop:
            {
              art->loop = resp->u.loop.loop;
              break;
            }
            case response_type_time:
            {
              audio_rt_seek_to(art, resp->u.time.time);
              break;
            }
            case response_type_read:
            {
              // free buffer, we didn't need it
              audio_rt_request_free(art, resp->u.read.bytes, resp->u.read.data);
              break;
            }
            case response_type_quit:
            {
              running = false;
              break;
            }
          }
          rres = ma_rb_commit_read(&art->response, rbytes, rbuffer);
        }
        else
        {
          rres = MA_NO_DATA_AVAILABLE;
          assert(! "split packet?");
        }
      }
      else
      {
        // wat do, error
      }
    }

    // playback
    double dt = AUDIO_HOP_SIZE / art->sample_rate;
    std::this_thread::sleep_for(std::chrono::nanoseconds((ma_int64) std::floor(1.0e9 * dt)));
    audio_rt_seek_to(art, art->time + art->speed * dt);

    // send playback position
    audio_rt_request_time(art, art->time);
  }
}

static void audio_rt_callback(ma_device* device, void* output, const void* input, ma_uint32 frames)
{
  struct audio_rt *art = (struct audio_rt *) device->pUserData;
  if (! art)
  {
    return;
  }

  // handle messages from decoder
  ma_result rres = MA_SUCCESS;
  while (rres == MA_SUCCESS && ma_rb_pointer_distance(&art->response) >= (ssize_t) sizeof(struct response))
  {
    size_t rbytes = sizeof(struct response);
    void *rbuffer = nullptr;
    rres = ma_rb_acquire_read(&art->response, &rbytes, &rbuffer);
    if (rres == MA_SUCCESS)
    {
      if (rbytes == sizeof(struct response))
      {
        struct response *resp = (struct response *) rbuffer;
        switch (resp->type)
        {
          case response_type_volume:
          {
            double volume = art->volume_target + art->volume_ramp * (art->volume_start - art->volume_target);
            art->volume_start = volume;
            art->volume_target = resp->u.volume.volume;
            art->volume_ramp = 1;
            break;
          }
          case response_type_speed:
          {
            //std::fprintf(stderr, "RT: speed := %f\n", resp->u.speed.speed);
            art->speed = resp->u.speed.speed; // FIXME TODO fade
            break;
          }
          case response_type_loop:
          {
            //std::fprintf(stderr, "RT: loop := %d\n", (int) resp->u.loop.loop);
            art->loop = resp->u.loop.loop;
            break;
          }
          case response_type_time:
          {
            //std::fprintf(stderr, "RT: seek to %.3f\n", resp->u.time.time);
            audio_rt_seek_to(art, resp->u.time.time);
            break;
          }
          case response_type_read:
          {
            //std::fprintf(stderr, "RT: read %ld %ld %ld %p\n", resp->u.read.start, resp->u.read.frames, resp->u.read.bytes, resp->u.read.data);
            // find corresponding target buffer
            bool stored = false;
            for (int b = 0; b < AUDIO_RT_BUFFERS; ++b)
            {
              if (art->buffer[b].start == resp->u.read.start && art->buffer[b].frames == resp->u.read.frames)
              {
                if (art->buffer[b].data)
                {
                  // already have buffer, assume data is the same, don't overwrite
                }
                else
                {
                  stored = true;
                  art->buffer[b].bytes = resp->u.read.bytes;
                  art->buffer[b].data = resp->u.read.data;
                  break;
                }
              }
            }
            if (! stored)
            {
              // free buffer, we didn't need it
              audio_rt_request_free(art, resp->u.read.bytes, resp->u.read.data);
              // FIXME leaks buffer memory if errored above
            }
            break;
          }
          case response_type_quit:
          {
            // not applicable to this rt mode
            break;
          }
        }
        rres = ma_rb_commit_read(&art->response, rbytes, rbuffer);
      }
      else
      {
        rres = MA_NO_DATA_AVAILABLE;
        assert(! "split packet?");
      }
    }
    else
    {
      // wat do, error
    }
  }

  // playback
  (void) input;
  float *out = (float *) output;
  double rms = 0;
  for (ma_uint32 n = 0; n < frames; ++n)
  {
    if ((art->frame % AUDIO_HOP_SIZE) == 0)
    {
//std::fprintf(stderr, "[ SHUNT\n");
      // shunt output buffers
//std::fprintf(stderr, "memmove %p %p %lld\n", art->outbuf, art->outbuf + art->channels * AUDIO_HOP_SIZE, sizeof(*art->outbuf) * art->channels * AUDIO_HOP_SIZE * (AUDIO_OVERLAP - 1));
      std::memmove(art->outbuf, art->outbuf + art->channels * AUDIO_HOP_SIZE, sizeof(*art->outbuf) * art->channels * AUDIO_HOP_SIZE * (AUDIO_OVERLAP - 1));
//std::fprintf(stderr, "memset p %d %lld\n", art->outbuf + art->channels * AUDIO_HOP_SIZE * (AUDIO_OVERLAP - 1), 0, sizeof(*art->outbuf) * art->channels * AUDIO_HOP_SIZE);
      std::memset(art->outbuf + art->channels * AUDIO_HOP_SIZE * (AUDIO_OVERLAP - 1), 0, sizeof(*art->outbuf) * art->channels * AUDIO_HOP_SIZE);
//std::fprintf(stderr, "] SHUNT\n", art->outbuf, art->outbuf + art->channels * AUDIO_HOP_SIZE, sizeof(*art->outbuf) * art->channels * AUDIO_HOP_SIZE * (AUDIO_OVERLAP - 1));
      ma_int64 current_frame = std::floor(art->time * art->sample_rate);
      for (int c = 0; c < art->channels; ++c)
      {
        // phase vocoder
        for (int o = 0; o < 2; ++o)
        {
          ma_int64 offset = o;
          if (art->speed < 0)
          {
            offset = -offset;
          }
          // deinterleave channel from buffer(s)
          for (int i = 0; i < AUDIO_FFT_SIZE; ++i)
          {
            ma_int64 frame = current_frame + i + offset;
            ma_int64 buffer_start = (frame / AUDIO_BLOCK_SIZE) * AUDIO_BLOCK_SIZE;
            int buffer = -1;
            for (int b = 0; b < AUDIO_RT_BUFFERS; ++b)
            {
              if (art->buffer[b].start == buffer_start)
              {
                buffer = b;
                break;
              }
            }
            ma_int64 j = frame - buffer_start;
            double sample = 0;
            if (0 <= buffer && buffer < AUDIO_RT_BUFFERS && 0 <= j && (ma_uint64) j < art->buffer[buffer].frames && art->buffer[buffer].data)
            {
              sample = art->buffer[buffer].data[j * art->channels + c];
            }
            else
            {
              art->errs_read_underrun++;
            }
            art->fft_in[i] = art->window[i] * sample;
          }
          // do FFT of source audio
          fftw_execute(art->fft[o]);
        }
        // compute phase advance
        for (int bin = 0; bin < AUDIO_FFT_SIZE; ++bin)
        {
          std::complex<double> then = to_std_complex(art->ifft_in[c][bin]);
          std::complex<double> now = to_std_complex(art->fft_out[0][bin]);
          std::complex<double> next = to_std_complex(art->fft_out[1][bin]);
          std::complex<double> advance = normalize(next / now);
          for (int power = 1; power < AUDIO_HOP_SIZE; power <<= 1)
          {
            advance *= advance;
          }
          then = normalize(then) * std::abs(now) * advance;
          to_fftw_complex(then, art->ifft_in[c][bin]);
        }
        // do IFFT
        fftw_execute(art->ifft[c]);
        // interleave channels to output buffers
        for (int i = 0; i < AUDIO_FFT_SIZE; ++i)
        {
          art->outbuf[i * art->channels + c] += art->window[i] * art->ifft_out[i];
        }
      }
      art->frame = 0;
    }
    // copy data
    double volume = art->volume_target + art->volume_ramp * (art->volume_start - art->volume_target);
    art->volume_ramp = std::max(0.0, art->volume_ramp - 1/0.01 / art->sample_rate); // 10ms
    for (int c = 0; c < art->channels; ++c)
    {
      out[n * art->channels + c] = volume * art->outbuf[art->frame * art->channels + c];
      rms += out[n * art->channels + c] * out[n * art->channels + c];
    }
    // advance
    art->frame++;
    // requests data from decoder as needed
    audio_rt_seek_to(art, art->time + art->speed / art->sample_rate);
  }
  rms /= frames;
  rms = std::sqrt(rms);
//std::fprintf(stderr, "RT: rms = %e (volume = %f, channels = %d)\n", rms, art->volume, art->channels);
  // send playback position
  audio_rt_request_time(art, art->time);
}

struct audio
{
  // non-realtime
  double total_time;
  double sample_rate;
  bool playing;
  bool playback_loop;
  double playback_speed;
  double playback_time;
  bool playback_mute;
  double playback_volume;
  ma_uint64 total_frames;
  bool have_decoder;
  ma_decoder_config decoder_config;
  ma_decoder decoder;
  bool have_device;
  ma_device_config device_config;
  ma_device device;
  ImGui::FileBrowser *browser;
  std::filesystem::path *soundtrack_file;
  // realtime
  bool have_rt;
  struct audio_rt art;
};

static ma_result audio_open(struct audio *au, const std::filesystem::path &file, double duration = 0);

extern struct audio *audio_new(void)
{
  struct audio *au = (struct audio *) calloc(1, sizeof(*au));
  if (! au)
  {
    return nullptr;
  }
  au->playback_volume = 1;
  au->playback_speed = 1;
  return au;
}

static void audio_close(struct audio *au)
{
  // close device
  if (au->have_device)
  {
    ma_device_uninit(&au->device); // stops device if necessary
    au->have_device = false;
  }
  // deinit rt
  if (au->have_rt)
  {
    audio_rt_uninit(&au->art);
    au->have_rt = false;
  }
  // close decoder
  if (au->have_decoder)
  {
    ma_decoder_uninit(&au->decoder);
    au->have_decoder = false;
  }
  au->total_frames = 0;
  au->sample_rate = 0;
  if (au->soundtrack_file)
  {
    delete au->soundtrack_file;
    au->soundtrack_file = nullptr;
  };
}

extern void audio_delete(struct audio *au)
{
  if (! au)
  {
    return;
  }
  audio_close(au);
  if (au->browser)
  {
    delete au->browser;
    au->browser = nullptr;
  }
}

static ma_result audio_playback_volume(struct audio *au, double volume);
static ma_result audio_playback_loop(struct audio *au, bool loop);

static ma_result audio_open(struct audio *au, const std::filesystem::path &file, double duration)
{
  // FIXME TODO only close/reopen device if channels/samplerate changed
  double now = au->playback_time;
  audio_close(au);
  ma_result res = MA_SUCCESS;
  if (file.empty())
  {
    // init rt for silence (0 channels)
    double sample_rate = 48000;
    au->total_frames = std::max(1.0, std::round(sample_rate * duration));
    res = audio_rt_init(&au->art, 0, sample_rate, au->total_frames);
    if (res != MA_SUCCESS)
    {
      audio_close(au);
      return res;
    }
    au->have_rt = true;
    au->soundtrack_file = new std::filesystem::path(file);
    res = MA_SUCCESS;
  }
  else
  {
    // open file decoder
    au->decoder_config = ma_decoder_config_init(ma_format_f32, 0, 0);
    res = ma_decoder_init_file(file.string().c_str(), &au->decoder_config, &au->decoder);
    if (res != MA_SUCCESS)
    {
      return res;
    }
    au->have_decoder = true;
    // init rt
    au->total_frames = ma_decoder_get_length_in_pcm_frames(&au->decoder);
    res = audio_rt_init(&au->art, au->decoder.outputChannels, au->decoder.outputSampleRate, au->total_frames);
    if (res != MA_SUCCESS)
    {
      audio_close(au);
      return res;
    }
    au->have_rt = true;
    // open audio device
    au->device_config = ma_device_config_init(ma_device_type_playback);
    au->device_config.playback.format = ma_format_f32;
    au->device_config.playback.channels = au->decoder.outputChannels;
    au->device_config.sampleRate = au->decoder.outputSampleRate;
    au->device_config.dataCallback = audio_rt_callback;
    au->device_config.pUserData = &au->art;
    res = ma_device_init(NULL, &au->device_config, &au->device);
    if (res != MA_SUCCESS)
    {
      audio_close(au);
      return res;
    }
    au->have_device = true;
    ma_device_start(&au->device);
    au->soundtrack_file = new std::filesystem::path(file);
    res = MA_SUCCESS;
  }
  if (res == MA_SUCCESS)
  {
    audio_set_time(au, now);
    audio_set_playing(au, audio_get_playing(au));
    audio_playback_loop(au, au->playback_loop);
    audio_set_speed(au, audio_get_speed(au));
    audio_playback_volume(au, au->playback_mute ? 0 : au->playback_volume);
  }
  return res;
}

extern bool audio_set_time(struct audio *au, double time)
{
  ma_result res = MA_NO_DATA_AVAILABLE;
  if (au->have_rt)
  {
    size_t bytes = sizeof(struct response);
    void *buffer = nullptr;
    res = ma_rb_acquire_write(&au->art.response, &bytes, &buffer);
    if (res == MA_SUCCESS)
    {
      if (bytes == sizeof(response))
      {
        struct response *resp = (struct response *) buffer;
        resp->type = response_type_time;
        resp->u.time.time = time;
        res = ma_rb_commit_write(&au->art.response, bytes, buffer);
      }
      else
      {
        res = MA_NO_DATA_AVAILABLE;
        assert(! "split packet?");
      }
    }
    if (res != MA_SUCCESS)
    {
      au->art.errs_response_overrun++;
    }
  }
  return res == MA_SUCCESS;
}

static ma_result audio_playback_speed(struct audio *au, double speed)
{
  ma_result res = MA_NO_DATA_AVAILABLE;
  if (au->have_rt)
  {
    size_t bytes = sizeof(struct response);
    void *buffer = nullptr;
    res = ma_rb_acquire_write(&au->art.response, &bytes, &buffer);
    if (res == MA_SUCCESS)
    {
      if (bytes == sizeof(struct response))
      {
        struct response *resp = (struct response *) buffer;
        resp->type = response_type_speed;
        resp->u.speed.speed = speed;
        res = ma_rb_commit_write(&au->art.response, bytes, buffer);
      }
      else
      {
        res = MA_NO_DATA_AVAILABLE;
        assert(! "split packet?");
      }
    }
    if (res != MA_SUCCESS)
    {
      au->art.errs_response_overrun++;
    }
  }
  return res;
}

static ma_result audio_playback_loop(struct audio *au, bool loop)
{
  ma_result res = MA_NO_DATA_AVAILABLE;
  if (au->have_rt)
  {
    size_t bytes = sizeof(struct response);
    void *buffer = nullptr;
    res = ma_rb_acquire_write(&au->art.response, &bytes, &buffer);
    if (res == MA_SUCCESS)
    {
      if (bytes == sizeof(struct response))
      {
        struct response *resp = (struct response *) buffer;
        resp->type = response_type_loop;
        resp->u.loop.loop = loop;
        res = ma_rb_commit_write(&au->art.response, bytes, buffer);
      }
      else
      {
        res = MA_NO_DATA_AVAILABLE;
        assert(! "split packet?");
      }
    }
    if (res != MA_SUCCESS)
    {
      au->art.errs_response_overrun++;
    }
  }
  return res;
}

static ma_result audio_playback_volume(struct audio *au, double volume)
{
  ma_result res = MA_NO_DATA_AVAILABLE;
  if (au->have_rt)
  {
    size_t bytes = sizeof(struct response);
    void *buffer = nullptr;
    res = ma_rb_acquire_write(&au->art.response, &bytes, &buffer);
    if (res == MA_SUCCESS)
    {
      if (bytes == sizeof(struct response))
      {
        struct response *resp = (struct response *) buffer;
        resp->type = response_type_volume;
        resp->u.volume.volume = volume;
        res = ma_rb_commit_write(&au->art.response, bytes, buffer);
      }
      else
      {
        res = MA_NO_DATA_AVAILABLE;
        assert(! "split packet?");
      }
    }
    if (res != MA_SUCCESS)
    {
      au->art.errs_response_overrun++;
    }
  }
  return res;
}

extern void audio_nrt(struct audio *au)
{
  // handle messages from audio rt
  ma_result rres = MA_SUCCESS;
  while (rres == MA_SUCCESS && ma_rb_pointer_distance(&au->art.request) >= (ssize_t) sizeof(struct request))
  {
    size_t rbytes = sizeof(struct request);
    void *rbuffer = nullptr;
    rres = ma_rb_acquire_read(&au->art.request, &rbytes, &rbuffer);
    if (rres == MA_SUCCESS)
    {
      if (rbytes == sizeof(struct request))
      {
        struct request *req = (struct request *) rbuffer;
        switch (req->type)
        {
          case request_type_free:
          {
            //std::fprintf(stderr, "NRT: free %ld %p\n", req->u.free.bytes, req->u.free.data);
            if (req->u.free.data)
            {
              std::free(req->u.free.data); // FIXME reuse buffers
            }
            break;
          }
          case request_type_read:
          {
            ma_int64 start = req->u.read.start;
            ma_uint64 frames = req->u.read.frames;
            //std::fprintf(stderr, "NRT: read %lld %lld\n", start, frames);
            size_t bytes = sizeof(float) * frames * au->art.channels;
            float *data = (float *) std::calloc(1, bytes); // FIXME reuse buffers
            if (start + (ma_int64) frames < 0 || (ma_int64) au->total_frames <= start)
            {
              //std::fprintf(stderr, "NRT: read out of range 0 %lld\n", au->total_frames);
              // entirely out of range, zero (nothing to do)
            }
            else if (0 <= start)
            {
              // entirely in range, read
              ma_result res = ma_decoder_seek_to_pcm_frame(&au->decoder, start);
              if (res == MA_SUCCESS)
              {
                 ma_uint64 rframes = ma_decoder_read_pcm_frames(&au->decoder, data, frames);
                 //std::fprintf(stderr, "NRT: read read %lld\n", rframes);
                 if (rframes < frames)
                 {
                   // reached EOF, but remainder is zero so no more to do
                 }
              }
              else
              {
                //std::fprintf(stderr, "NRT: seek failed %d\n", res);
              }
            }
            else
            {
              // block overlaps 0, should never happen?
              //std::fprintf(stderr, "NRT: read overlaps 0\n");
              ma_result res = ma_decoder_seek_to_pcm_frame(&au->decoder, 0);
              if (res == MA_SUCCESS)
              {
                 ma_uint64 rframes = ma_decoder_read_pcm_frames(&au->decoder, data + (-start) * au->art.channels, frames - (-start));
                 //std::fprintf(stderr, "NRT: read read %lld\n", rframes);
                 if (rframes < frames - (-start))
                 {
                   // reached EOF, but remainder is zero so no more to do
                 }
              }
              else
              {
                //std::fprintf(stderr, "NRT: seek failed %d\n", res);
              }
            }
            audio_rt_response_read(&au->art, start, frames, bytes, data);
            break;
          }
          case request_type_time:
          {
            //std::fprintf(stderr, "NRT: time %f\n", req->u.time.time);
            au->playback_time = req->u.time.time;
            break;
          }
        }
      }
      else
      {
        // wat do, got a split packet?
        rres = MA_NO_DATA_AVAILABLE;
        //std::fprintf(stderr, "ERROR: split packet: wanted %d got %d\n", sizeof(struct request), rbytes);
        assert(! "split packet?");
      }
      rres = ma_rb_commit_read(&au->art.request, rbytes, rbuffer);
    }
  }
}

extern bool audio_display_early(struct audio *au)
{
  bool new_soundtrack = false;
  if (! au->browser)
  {
    au->browser = new ImGui::FileBrowser(ImGuiFileBrowserFlags_CloseOnEsc);
    au->browser->SetTitle("Soundtrack Audio File");
    au->browser->SetTypeFilters({ ".wav", ".flac", ".ogg", ".mp3" });
  }
  ImGui::Begin("Audio", nullptr, ImGuiWindowFlags_HorizontalScrollbar);
  bool red = ! au->have_rt;
  if (red)
  {
    ImGui::PushStyleColor(ImGuiCol_Button, button_red);
  }
  bool soundtrack_is_empty = true;
  if (au->soundtrack_file && ! au->soundtrack_file->empty())
  {
    soundtrack_is_empty = false;
  }
  int audio_preset = soundtrack_is_empty ? 0 : 1;
  const char *audio_presets[] = { "Silence", "Soundtrack" };
  ImGui::PushItemWidth(100);
  bool audio_preset_changed = false;
  double seconds = 0;
  if (au->have_rt)
  {
    seconds = au->art.duration / au->art.sample_rate;
  }
  if (ImGui::Combo("##Audio", &audio_preset, audio_presets, IM_ARRAYSIZE(audio_presets)))
  {
    if (audio_preset == 1)
    {
      au->browser->Open();
    }
    audio_preset_changed = true;
  }
  if (red)
  {
    ImGui::PopStyleColor(1);
  }
  ImGui::PopItemWidth();
  if (ImGui::IsItemHovered())
  {
    ImGui::SetTooltip("Set audio source.");
  }
  if (audio_preset == 0)
  {
    ImGui::SameLine();
    ImGui::PushItemWidth(200);
    if (ImGui::InputDouble("Duration", &seconds, 1, 60, "%.3fs") || audio_preset_changed)
    {
      if (seconds > 0)
      {
        std::filesystem::path silence;
        audio_open(au, silence, seconds);
        new_soundtrack = true;
      }
    }
    ImGui::PopItemWidth();
    if (ImGui::IsItemHovered())
    {
      ImGui::SetTooltip("Set duration in seconds.");
    }
  }
  if (audio_preset == 1 && au->soundtrack_file)
  {
    ImGui::SameLine();
    ImGui::Text("%s", au->soundtrack_file->string().c_str());
  }
  if (audio_preset == 1 && au->have_rt)
  {
    ImGui::Text("Sample Rate: %.0fHz", au->art.sample_rate);
    ImGui::SameLine();
    ImGui::Text("Channels: %d", au->art.channels);
    ImGui::SameLine();
    ImGui::Text("Duration: %.3fs", au->art.duration / au->art.sample_rate);
  }
  if (au->have_rt)
  {
    float seek = au->playback_time;
    if (ImGui::SliderFloat("Time", &seek, 0, au->art.duration / au->art.sample_rate, "%.3fs"))
    {
      audio_set_time(au, seek);
    }
    if (ImGui::IsItemHovered())
    {
      ImGui::SetTooltip("Seek to time in soundtrack.");
    }
    if (ImGui::Button("Rewind"))
    {
      audio_set_time(au, 0);
    }
    if (ImGui::IsItemHovered())
    {
      ImGui::SetTooltip("Seek to the start of the soundtrack.");
    }
    ImGui::SameLine();
    if (ImGui::Button("Stop"))
    {
      au->playing = false;
      audio_playback_speed(au, 0);
    }
    if (ImGui::IsItemHovered())
    {
      ImGui::SetTooltip("Disable soundtrack play head advance.");
    }
    ImGui::SameLine();
    if (ImGui::Button("Play"))
    {
      au->playing = true;
      audio_playback_speed(au, au->playback_speed);
    }
    if (ImGui::IsItemHovered())
    {
      ImGui::SetTooltip("Enable soundtrack play head advance.");
    }
    ImGui::SameLine();
    if (ImGui::Checkbox("Loop", &au->playback_loop))
    {
      audio_playback_loop(au, au->playback_loop);
    }
    if (ImGui::IsItemHovered())
    {
      ImGui::SetTooltip("Enable soundtrack play head loop around ends.");
    }
    ImGui::SameLine();
    if (ImGui::Button("1x"))
    {
      au->playback_speed = 1;
      if (au->playing)
      {
        audio_playback_speed(au, au->playback_speed);
      }
    }
    if (ImGui::IsItemHovered())
    {
      ImGui::SetTooltip("Set playback speed to 1x (normal speed).");
    }
    ImGui::SameLine();
    float speed = au->playback_speed;
    ImGui::PushItemWidth(100);
    if (ImGui::SliderFloat("##Speed", &speed, -16, 16, "%.3fx"))
    {
      au->playback_speed = speed;
      if (au->playing)
      {
        audio_playback_speed(au, au->playback_speed);
      }
    }
    if (ImGui::IsItemHovered())
    {
      ImGui::SetTooltip("Set playback speed (negative plays backwards).");
    }
    ImGui::PopItemWidth();
    ImGui::SameLine();
    if (ImGui::Checkbox("##Mute", &au->playback_mute))
    {
      audio_playback_volume(au, au->playback_mute ? 0 : au->playback_volume);
    }
    if (ImGui::IsItemHovered())
    {
      ImGui::SetTooltip("Mute audio (set volume to -inf dB, 0.0).");
    }
    ImGui::SameLine();
    ImGui::PushItemWidth(100);
    float volume = rmstodbfs(au->playback_volume);
    if (ImGui::SliderFloat("##Volume", &volume, -60, 10, "%.1fdB"))
    {
      au->playback_mute = false;
      au->playback_volume = dbfstorms(volume);
      audio_playback_volume(au, au->playback_mute ? 0 : au->playback_volume);
    }
    if (ImGui::IsItemHovered())
    {
      ImGui::SetTooltip("Set playback volume in dB (0 is full scale).");
    }
    ImGui::PopItemWidth();
    ma_uint64 errs_request_overrun = au->art.errs_request_overrun;
    ma_uint64 errs_response_overrun = au->art.errs_response_overrun;
    ma_uint64 errs_read_underrun = au->art.errs_read_underrun;
    if (errs_request_overrun || errs_response_overrun || errs_read_underrun)
    {
      ImGui::Text("Audio Errors: %u / %u / %u", (unsigned int) errs_request_overrun, (unsigned int) errs_response_overrun, (unsigned int) errs_read_underrun);
      ImGui::SameLine();
      if (ImGui::Button("Reset"))
      {
        au->art.errs_request_overrun = 0;
        au->art.errs_response_overrun = 0;
        au->art.errs_read_underrun = 0;
      }
    }
  }
  ImGui::End();
  return new_soundtrack;
}

extern bool audio_display_late(struct audio *au)
{
  if (! au)
  {
    return false;
  }
  if (! au->browser)
  {
    return false;
  }
  au->browser->Display();
  if (au->browser->HasSelected())
  {
    audio_open(au, au->browser->GetSelected());
    au->browser->ClearSelected();
    if (au->have_rt)
    {
      return true;
    }
  }
  return false;
}

extern double audio_get_duration(const struct audio *au)
{
  if (! au)
  {
    return 0;
  }
  if (! au->have_rt)
  {
    return 0;
  }
  return au->art.duration / au->art.sample_rate;
}

extern double audio_get_time(const struct audio *au)
{
  if (! au)
  {
    return 0;
  }
  if (! au->have_rt)
  {
    return 0;
  }
  return au->playback_time;
}

extern bool audio_set_playing(struct audio *au, bool playing)
{
  if (au)
  {
    au->playing = playing;
    return MA_SUCCESS == audio_playback_speed(au, au->playing ? au->playback_speed : 0);
  }
  return false;
}

extern bool audio_get_playing(const struct audio *au)
{
  return au->playing;
}

extern bool audio_set_speed(struct audio *au, double speed)
{
  if (au)
  {
    au->playback_speed = speed;
    return MA_SUCCESS == audio_playback_speed(au, au->playing ? au->playback_speed : 0);
  }
  return false;
}

extern double audio_get_speed(const struct audio *au)
{
  return au->playback_speed;
}

extern std::filesystem::path audio_get_soundtrack(const struct audio *au)
{
  if (au)
  {
    if (au->soundtrack_file)
    {
      return *au->soundtrack_file;
    }
  }
  return std::filesystem::path();
}

extern enum audio_type audio_get_type(const struct audio *au)
{
  if (au)
  {
    if (au->soundtrack_file)
    {
      if (! au->soundtrack_file->empty())
      {
        return audio_type_soundtrack;
      }
    }
  }
  return audio_type_silence;
}

extern bool audio_set_volume(struct audio *au, double volume)
{
  if (au)
  {
    au->playback_volume = volume;
    return MA_SUCCESS == audio_playback_volume(au, au->playback_mute ? 0 : au->playback_volume);
  }
  return false;
}

extern double audio_get_volume(const struct audio *au)
{
  return au->playback_volume;
}

extern bool audio_set_mute(struct audio *au, bool mute)
{
  if (au)
  {
    au->playback_mute = mute;
    return MA_SUCCESS == audio_playback_volume(au, au->playback_mute ? 0 : au->playback_volume);
  }
  return false;
}

extern bool audio_get_mute(const struct audio *au)
{
  return au->playback_mute;
}

extern bool audio_set_loop(struct audio *au, bool loop)
{
  if (au)
  {
    au->playback_loop = loop;
    return MA_SUCCESS == audio_playback_loop(au, au->playback_loop);
  }
  return false;
}

extern bool audio_get_loop(const struct audio *au)
{
  return au->playback_loop;
}

extern bool audio_set_soundtrack(struct audio *au, const std::filesystem::path &file)
{
  return MA_SUCCESS == audio_open(au, file);
}

extern bool audio_set_silence(struct audio *au, double duration)
{
  return MA_SUCCESS == audio_open(au, std::filesystem::path(), duration);
}

#include "extras/stb_vorbis.c"
