// zoomasm -- zoom video assembler
// (c) 2019,2020 Claude Heiland-Allen
// SPDX-License-Identifier: AGPL-3.0-or-later

#include "process.h"

#include <sstream>
#include <cstring>
#include <cstdlib>

#include <windows.h>

#include "utility.h"

struct process
{
  char *command_line;
  SECURITY_ATTRIBUTES attributes;
  HANDLE stdin_read;
  HANDLE stdin_write;
  HANDLE stdout_read;
  HANDLE stdout_write;
  HANDLE stderr_read;
  HANDLE stderr_write;
  PROCESS_INFORMATION information;
  STARTUPINFO startup;
};

static std::string quote(std::string &arg)
{
  std::ostringstream o;
  bool backslash = false;
  if (arg.find(' ') != std::string::npos || arg.find('\t') != std::string::npos)
  {
    o << '"';
  }
  for (auto c : arg)
  {
    if (c == '"')
    {
      if (backslash)
      {
        o << '\\';
        o << '\\';
      }
      else
      {
        o << '\\';
      }
    }
    o << c;
    backslash = c == '\\';
  }
  if (arg.find(' ') != std::string::npos || arg.find('\t') != std::string::npos)
  {
    if (backslash)
    {
      o << '\\';
    }
    o << '"';
  }
  return o.str();
}

extern struct process *process_open(const std::vector<std::string> &args)
{
  struct process *proc = (struct process *) std::calloc(1, sizeof *proc);
  if (proc)
  {
  std::ostringstream command_line;
  bool first = true;
  for (auto arg : args)
  {
    if (! first)
    {
      command_line << " ";
    }
    first = false;
    command_line << quote(arg);
  }
  proc->command_line = mystrdup(command_line.str().c_str());
  if (proc->command_line)
  {
  proc->attributes.nLength = sizeof(proc->attributes);
  proc->attributes.bInheritHandle = TRUE;
  // stdin
  if (CreatePipe(&proc->stdin_read, &proc->stdin_write, &proc->attributes, 0))
  {
  if (SetHandleInformation(proc->stdin_write, HANDLE_FLAG_INHERIT, 0))
  {
  // stdout
  if (CreatePipe(&proc->stdout_read, &proc->stdout_write, &proc->attributes, 0))
  {
  if (SetHandleInformation(proc->stdout_read, HANDLE_FLAG_INHERIT, 0))
  {
  // stderr
  if (CreatePipe(&proc->stderr_read, &proc->stderr_write, &proc->attributes, 0))
  {
  if (SetHandleInformation(proc->stderr_read, HANDLE_FLAG_INHERIT, 0))
  {
  // startup information
  proc->startup.cb = sizeof(proc->startup);
  proc->startup.hStdError = proc->stderr_write;
  proc->startup.hStdOutput = proc->stdout_write;
  proc->startup.hStdInput = proc->stdin_read;
  proc->startup.dwFlags |= STARTF_USESTDHANDLES;
  // create process
  if (CreateProcess(NULL, proc->command_line, NULL, NULL, TRUE, 0, NULL, NULL, &proc->startup, &proc->information))
  {
    // close the pipe ends not needed by the parent, to detect child exit
    CloseHandle(proc->stderr_write); proc->stderr_write = INVALID_HANDLE_VALUE;
    CloseHandle(proc->stdout_write); proc->stdout_write = INVALID_HANDLE_VALUE;
    CloseHandle(proc->stdin_read); proc->stdin_read = INVALID_HANDLE_VALUE;
    return proc;
  }
  else
  {
    // CreateProcess failed
  }
  }
  else
  {
    // SetHandleInformation stderr failed
    CloseHandle(proc->stderr_read); proc->stderr_read = INVALID_HANDLE_VALUE;
    CloseHandle(proc->stderr_write); proc->stderr_write = INVALID_HANDLE_VALUE;
  }
  }
  else
  {
    // CreatePipe stderr failed
  }
  }
  else
  {
    // SetHandleInformation stdout failed
    CloseHandle(proc->stdout_read); proc->stdout_read = INVALID_HANDLE_VALUE;
    CloseHandle(proc->stdout_write); proc->stdout_write = INVALID_HANDLE_VALUE;
  }
  }
  else
  {
    // CreatePipe stdout failed
  }
  }
  else
  {
    // SetHandleInformation stdin failed
    CloseHandle(proc->stdin_read); proc->stdin_read = INVALID_HANDLE_VALUE;
    CloseHandle(proc->stdin_write); proc->stdin_write = INVALID_HANDLE_VALUE;
  }
  }
  else
  {
    // CreatePipe stdin failed
  }
  }
  else
  {
    // strdup failed
    std::free(proc);
  }
  }
  else
  {
    // calloc failed
  }
  return nullptr;
}

extern void process_close(struct process *proc)
{
  CloseHandle(proc->stderr_read); proc->stderr_read = INVALID_HANDLE_VALUE;
  CloseHandle(proc->stdout_read); proc->stdout_read = INVALID_HANDLE_VALUE;
  CloseHandle(proc->stdin_write); proc->stdin_write = INVALID_HANDLE_VALUE;
  WaitForSingleObject(proc->information.hProcess, INFINITE);
  CloseHandle(proc->information.hProcess);
  CloseHandle(proc->information.hThread);
  free(proc->command_line);
  free(proc);
}

extern void process_stdin_eof(struct process *proc)
{
  if (CloseHandle(proc->stdin_write))
  {
    proc->stdin_write = INVALID_HANDLE_VALUE;
  }
  else
  {
    // error
  }
}

extern ssize_t process_stdin(struct process *proc, const void *data, ssize_t bytes)
{
  DWORD written = 0;
  BOOL success = WriteFile(proc->stdin_write, data, bytes, &written, NULL);
  if (success)
  {
    return written;
  }
  return -1;
}

extern ssize_t process_stdout(struct process *proc, void *data, ssize_t bytes)
{
  DWORD read_bytes = 0;
  BOOL success = ReadFile(proc->stdout_read, data, bytes, &read_bytes, NULL);
  if (success)
  {
    return read_bytes;
  }
  else
  {
    return -1;
  }
}

extern ssize_t process_stderr(struct process *proc, void *data, ssize_t bytes)
{
  DWORD read_bytes = 0;
  BOOL success = ReadFile(proc->stderr_read, data, bytes, &read_bytes, NULL);
  if (success)
  {
    return read_bytes;
  }
  else
  {
    return -1;
  }
}
